import asyncio
import asyncpg
import datetime
import json
from dl2050utils.core import *
from dl2050utils.log import BaseLog

def run_sql_file(db, path):
    with open(path, 'r') as f: Q_CREATE=f.readlines()
    cmds = ' '.join(Q_CREATE).split(';')
    for q in cmds:
        if len(q)<10: continue
        db.sync_execute(q+';')

def run_slq_cmds(db, q):
    for q in q.split(';'):
        if len(q)<10: continue
        db.sync_execute(q+';')

def db_insert_df(db, tbl, df, dmap=None):
    if dmap is None:
        dmap={c:c for c in df.columns}
    else:
        df = df.rename(columns=dmap)
    cols = [dmap[c] for c in dmap]
    df = df[cols]
    for i in range(len(df)):
        row = df.iloc[i]
        d = {c:row[c] for c in cols}
        if db.sync_insert(tbl, d):
            print('DB insert error')
    return False

async def calc_query_rows(con, q):
    # db.sync_select('pg_class', cols=['reltuples'], filters={'relname': 'diagsg'})
    res = await con.fetchrow(f'explain(format json) {q}')
    if res is None:
        return None
    res = json.loads(res['QUERY PLAN'])
    if res is None or not len(res):
        return None
    return oget(res[0],['Plan','Plan Rows'])

def fix_types(d):
    for k in d.keys():
        if isinstance(d[k], str): d[k]=d[k].strip()
        if isinstance(d[k], datetime.date):
            d[k] = d[k] if isinstance(d[k], datetime.datetime) else d[k].strftime("%Y-%m-%d")
    return d

def strip(e):
    if type(e)!=str: return e
    e = e.replace('\'', '')
    e.replace('\"', '')
    e.replace('\n', ' ')
    return e

def get_repr(e):
    if e is None: return 'NULL'
    if type(e)==list:
        items = [f'"{str(e1)}"' for e1 in e]
        return f"'{{{' ,'.join(items)}}}'"
    if type(e)==dict: return f"'{json.dumps(e)}'"
    if type(e)==str or type(e)==datetime.datetime: return f"'{strip(e)}'"
    return f"{e}"

def parse_filters(fs):
    if fs is None: return []
    if type(fs)!=dict and type(fs)!=list: return []
    if type(fs)==list: return [e for e in fs if type(e)==dict and 'col' in e and 'val in e']
    return [{'col':e, 'val':fs[e]} for e in fs]

class DB():
    def __init__(self, cfg=None, log=None, dbname=None):
        if cfg is None: cfg={'db':{'host':'db','port':5432,'user':'postgres','passwd':'rootroot','dbname':'postgres'}}
        self.cfg, self.LOG = cfg, log or BaseLog()
        host = oget(cfg, ['db','host'], 'db')
        port = oget(cfg, ['db','port'], 5432)
        user = oget(cfg, ['db','user'], 'postgres')
        passwd = oget(cfg, ['db','passwd'], 'rootroot')
        if dbname is None:
            dbname = oget(cfg, ['db', 'dbname'], 'postgres')
        self.url = f'postgres://{user}:{passwd}@{host}:{port}/{dbname}'
        self.dbname = dbname

    async def startup(self, min_size=5, max_size=20, loop=None):
        try:
            self.pool = await asyncpg.create_pool(self.url, min_size=min_size, max_size=max_size, loop=loop)
        except Exception as e:
            self.LOG(4, 0, label='DBPG', label2='startup', msg=str(e))
            return True
        self.LOG(2, 0, label='DBPG', label2='startup', msg=f'CONNECTED POOL to {self.dbname}')
        return False

    def shutdown(self):
        self.pool.terminate()
        self.LOG(2, 0, label='DBPG', label2='shutdown', msg='DISCONNECTED')
        return False
    
    async def execute(self, q):
        con = await self.pool.acquire()
        try:
            res = await con.execute(q)
        except Exception as err:
            self.LOG(4, 0, label='DBPG', label2='execute', msg={'error_msg': str(err), 'query': q})
            return None
        finally:
            await self.pool.release(con)
        return res

    async def query(self, q, one=False, nrows=False, offset=None, limit=None):
        con = await self.pool.acquire()
        if q[:6].upper()!='SELECT': one=True # accounts for returning in insert
        if one: nrows=False
        try:
            if(one):
                res = await con.fetchrow(q)
                if res is None:
                    await self.pool.release(con)
                    return None
            else:
                if nrows:
                    nr = await calc_query_rows(con, q)
                    if nr is None:
                        self.LOG(4, 0, label='DB', label2='calc_query_rows', msg=q)
                        await self.pool.release(con)
                        return None
                if offset or limit: q=q[:-1]
                if offset is not None: q += f" OFFSET {offset}"
                if limit is not None: q += f" LIMIT {limit}"
                res = await con.fetch(q)
        except Exception as e:
            self.LOG(4, 0, label='DB', label2='query', msg=str(e))
            return None
        finally:
            await self.pool.release(con)
        if res is None:
            return None
        if(one):
            return fix_types(dict(res))
        res = [fix_types(dict(row)) for row in res[:100000]]
        if nrows:
            return {'data':res, 'nrows':nr}
        return res
    
    async def get_trows(self, tbl):
        q = f"select reltuples as nrows from pg_class where relname='{tbl}'"
        con = await self.pool.acquire()
        res = await con.fetchrow(q)
        if res is None:
            return -1
        return int(dict(res)['nrows'])

    async def select(self, tbl, filters=None, sfilters=None, cols='*', sort=None, ascending=True, offset=None, limit=None, one=False):
        if cols is not None: cols=', '.join(cols)
        offset,limit = offset or 0,limit or 32
        q = f"SELECT {cols} FROM {tbl}"
        fs,sfs = parse_filters(filters),parse_filters(sfilters)
        fs = [{**e, 'op':'='} if 'op' not in e else e for e in fs]
        if len(fs):
            fs = ' AND '.join([f"{e['col']}{e['op']}'{e['val']}'" for e in fs])
            q += f" WHERE {fs}"
        if len(sfs):
            q+=' WHERE ' if not len(fs) else ' AND '
            q += ' AND '.join([f"{e['col']} ILIKE '%{e['val']}%'" for e in sfs])
        if sort is not None: q += f" ORDER BY {sort} " + ("ASC" if ascending else "DESC")
        q += ';'
        return await self.query(q, one=one, nrows=True, offset=offset, limit=limit)

    async def select_one(self, tbl, filters):
        return await self.select(tbl, filters=filters, one=True)
    
    async def insert(self, tbl, d, return_key=None):
        q = f"INSERT INTO {tbl} ("
        for k in d.keys(): q += f"{k}, "
        q = q[:-2] + ") VALUES ("
        for k in d.keys(): q = q + get_repr(d[k]) + ', '
        q = q[:-2] + ")"
        if return_key is not None:
            q += f" returning {return_key};"
            res = await self.query(q, one=True)
            if res is not None and len(res):
                return res[return_key]
            return None
        res = await self.execute(q)
        if res=='INSERT 0 1': return False
        return True

    async def delete(self, tbl, k, v):
        if k is None or v is None: return True
        q = f"DELETE FROM {tbl} WHERE {k}='{v}'"
        res = await self.execute(q)
        if res is None: return True
        n = int(res[7:])
        if n==0: return True
        return False
    
    async def delete_all(self, tbl):
        q = f"DELETE FROM {tbl}"
        res = await self.execute(q)
        if res is None: return True
        return False

    async def update(self, tbl, ks, d):
        ks = listify(ks)
        for k in ks:
            if not k in d:
                self.LOG(4, 0, label='DBPG', label2='update', msg=f'key error: {k}')
                return True
        setvars = ', '.join([f"{k}={get_repr(d[k])}" for k in d if k not in ks+['nrows'] and d[k] is not None])
        q = f"UPDATE {tbl} SET {setvars}"
        filters = ' AND '.join([f"{k}='{d[k]}'" for k in ks if d[k] is not None])
        q += f" WHERE {filters};"
        res = await self.execute(q)
        if res is None: return True
        n = int(res[7:])
        if n==0: return True
        return False

    async def insert_rows(self, tbl, rows, delete=False):
        if delete:
            res = await self.execute(f'DELETE FROM {tbl}')
            if res is None:
                return res
        for row in rows:
            res = await self.insert(tbl, row)
            if res is None:
                return res
        return len(rows)

    def sync_startup(self, *args, **kwargs): return asyncio.get_event_loop().run_until_complete(self.startup(*args, **kwargs))
    def sync_execute(self, q): return asyncio.get_event_loop().run_until_complete(self.execute(q))
    def sync_query(self, *args, **kwargs): return asyncio.get_event_loop().run_until_complete(self.query(*args, **kwargs))
    def sync_select(self, *args, **kwargs): return asyncio.get_event_loop().run_until_complete(self.select(*args, **kwargs))
    def sync_select_one(self, *args, **kwargs): return asyncio.get_event_loop().run_until_complete(self.select_one(*args, **kwargs))
    def sync_insert(self, *args, **kwargs): return asyncio.get_event_loop().run_until_complete(self.insert(*args, **kwargs))
    def sync_delete(self, *args, **kwargs): return asyncio.get_event_loop().run_until_complete(self.delete(*args, **kwargs))
    def sync_delete_all(self, *args, **kwargs): return asyncio.get_event_loop().run_until_complete(self.delete_all(*args, **kwargs))
    def sync_update(self,  *args, **kwargs): return asyncio.get_event_loop().run_until_complete(self.update(*args, **kwargs))



import time
from datetime import datetime
import json
from aio_pika import connect_robust, Message, DeliveryMode
from dl2050utils.core import listify

JOB_CREATE = 0
JOB_START = 1
JOB_DONE = 2
JOB_NOTIFY = 3
JOB_DELIVER = 4
JOB_ERROR = 99

class MQ():
    def __init__(self, log, db, qnames, user='admin', passwd='password'):
        self.LOG,self.db,self.qnames = log,db,listify(qnames)
        self.url = f'amqp://{user}:{passwd}@mq:5672/'

    async def startup(self, loop=None):
        try:
            self.con = await connect_robust(self.url, loop=loop)
            self.ch = await self.con.channel()
            for qname in self.qnames:
                await self.ch.declare_queue(qname, durable=True, auto_delete=False)
        except Exception as e:
            self.LOG(4, 0, label='MQ', label2='connect', msg=str(e))
            return True
        self.LOG(2, 0, label='MQ', label2='STARTUP', msg='OK')
        return False

    def sync_startup(self, loop=None):
        return loop.run_until_complete(self.startup(loop=loop))
    
    async def consumer(self, w, qname, cb):
        if qname not in self.qnames:
            self.LOG(4, 0, label='MQ', label2='EXCEPTION', msg=f'Invalid qname {qname}')
            raise f'Invalid qname {qname}'
        q = await self.ch.declare_queue(qname, durable=True, auto_delete=False)
        async with q.iterator() as it:
            async for msg in it:
                async with msg.process():
                    await self.process_msg(w, msg, cb)

    async def process_msg(self, w, msg, cb):
        t0 = time.time()
        payload = json.loads(msg.body.decode())
        uid,jid = payload['uid'],payload['jid']
        self.LOG(2, 0, label='MQ', label2=f'JOB SCHEDULED', msg={'uid':uid, 'jid':jid})
        try:
            err = cb(w, jid, payload)
        except Exception as exc:
                self.LOG(4, time.time()-t0, label='MQ', label2=f'JOB EXCEPTION', msg={'uid':uid, 'jid':jid, 'exception':str(exc)})
                await self.job_error(jid)
                return
        if err:
            await self.mq.job_error(jid)
            self.LOG(4, time.time()-t0, label='MQ', label2=f'JOB ERROR', msg={'uid':uid ,'jid':jid})
            self.job_error(jid)
            return
        await self.job_done(jid)
        # msg.ch.basic_ack(delivery_tag = msg.method.delivery_tag)
        self.LOG(2, time.time()-t0, label='MQ', label2=f'JOB EXECUTED', msg={'uid':uid, 'jid':jid})

    async def publish(self, qname, uid, payload):
        job = {'uid':uid, 'payload':payload, 'jstatus':JOB_CREATE, 'eta':0, 'ts_create':datetime.now()}
        jid = await self.db.insert('jobs', job, return_key='jid')
        if jid is None: return None
        try:
            payload['uid'],payload['jid'] = uid,jid
            msg = Message(body=json.dumps(payload).encode(), delivery_mode=DeliveryMode.PERSISTENT)
            await self.ch.default_exchange.publish(msg, routing_key=qname)
        except Exception as e:
            self.LOG(4, 0, label='MQ', label2='publish', msg=str(e))
            return None
        return jid
        
    async def job_update(self, jid, status=None, eta=None, result=None):
        job = {'jid': jid}
        if status is not None:
            job['jstatus'] = status
            if status==JOB_START: job['ts_start'] = datetime.now()
            if status==JOB_DONE: job['ts_done'] = datetime.now()
        if eta is not None: job['eta']=eta
        if result is not None: job['result']=result
        self.LOG(1, 0, label='MQ', label2='job_update', msg=job)
        res = await self.db.update('jobs', 'jid', job)
        if res is None or res==0: return True
        return False

    async def job_start(self, jid, eta=None): return await self.job_update(jid, status=JOB_START, eta=eta)
    async def job_update_eta(self, jid, eta): return await self.job_update(jid, eta=eta)
    async def job_done(self, jid): return await self.job_update(jid, status=JOB_DONE)
    async def job_notify(self, jid): return await self.job_update(jid, status=JOB_NOTIFY)
    async def job_deliver(self, jid): return await self.job_update(jid, status=JOB_DELIVER)
    async def job_result(self, jid, result): return await self.job_update(jid, result=result)
    async def job_error(self, jid): return await self.job_update(jid, status=JOB_ERROR, eta=0)
    async def get_jobs(self, d): return await self.db.select('jobs', d)
    async def get_pending_jobs(self, uid): return await self.db.select('jobs', {'uid':uid,'jstatus':1})
    async def get_job(self, jid): return await self.db.select_one('jobs', {'jid':jid})



import time
from pathlib import Path
from starlette.responses import FileResponse
from starlette.routing import Route
from starlette.authentication import requires
from dl2050utils.core import oget
from dl2050utils.restutils import HTTPException, rest_ok, rest_error, get_meta, get_required_args,\
                                  get_required, get_optional_args, mk_key, get_hash

class App():

    def __init__(self, cfg, LOG, NOTIFY, db, mq, auth, path, routes=[], appstartup=None, perm=None):
        self.cfg,self.LOG,self.NOTIFY,self.db,self.mq,self.auth = cfg,LOG,NOTIFY,db,mq,auth
        self.routes,self.appstartup,self.perm = routes,appstartup,perm
        self.d = {'LOG':LOG, 'path':path, 'db':db, 'mq':mq}

    async def startup(self):
        model,meta = oget(self.cfg,['app','model']),None
        if model is not None: meta = await get_meta(self.db, model)
        self.d['meta'] = meta
        if self.appstartup is None: return False
        return await self.appstartup(self.d)

    def shutdown(self):
        self.LOG(2, 0, label='APP', label2='shutdown', msg='OK')
        return False   

    def get_routes(self):
        BASE_ROUTES = [
            Route('/api/get_meta', endpoint=self.get_meta, methods=['GET']),
            Route('/api/public2', endpoint=self.public2, methods=['GET']),
            Route('/api/private', endpoint=self.private, methods=['GET']),
            Route('/api/download/{fname}', endpoint=self.download, methods=['GET']),
            Route('/api/upload/{size}', endpoint=self.upload, methods=['GET']),
            Route('/api/publish_job', endpoint=self.publish_job, methods=['POST']),
            Route('/api/get_jobs', endpoint=self.get_jobs, methods=['POST']),
            Route('/api/get_pending_jobs', endpoint=self.get_pending_jobs, methods=['POST']),
            Route('/api/get_job', endpoint=self.get_job, methods=['POST']),
        ]
        APP_ROUTES = [Route(e, endpoint=self.app_route, methods=['POST']) for e in self.routes]
        return BASE_ROUTES + APP_ROUTES

    @requires('authenticated')
    async def get_meta(self, request):
        u = await self.auth.check_auth(request)
        return rest_ok(self.d['meta'])

    @requires('authenticated')
    async def public2(self, request):
        u = await self.auth.path_params(request)
        fname = request.path_params['fname']
        path = self.d['path']
        p = Path(f'{path}/public2/{fname}')
        return FileResponse(str(p), filename=fname)

    @requires('authenticated')
    async def private(self, request):
        u = await self.auth.check_auth(request)
        uid = u['uid']
        fname = request.path_params['fname']
        path = self.d['path']
        p = Path(f'{path}/private/{uid}/{fname}')
        return FileResponse(str(p), media_type='application/vnd.ms-excel', filename=fname)

    @requires('authenticated')
    async def download(self, request):
        u = await self.auth.check_auth(request)
        uid = u['uid']
        fname = request.path_params['fname']
        path = self.d['path']
        p = Path(f'{path}/tmp/{int(time.time())//86400}-{uid}-{fname}')
        if not p.is_file():
            raise HTTPException(404, detail=f'File not found')
        return FileResponse(str(p), media_type='application/vnd.ms-excel', filename=fname)

    @requires('authenticated')
    async def upload(self, request):
        u = await self.auth.check_auth(request)
        size = request.path_params['size']
        self.perm(self.d, u, request.url.path, {'size':size})
        res = {'key':mk_key(16), 'size':size}
        secret = oget(self.cfg,['app','upload','secret'])
        if not secret:
            return rest_error(self.LOG, 'APP', 'upload', 'Upload secret not defined')
        h = get_hash(res, secret)
        res['h'] = h
        return rest_ok(res)

    @requires('authenticated')
    async def publish_job(self, request):
        u = await self.auth.check_auth(request)
        uid = u['uid']
        data = await request.json()
        args = get_required_args(data,['q','payload'])
        q,payload = args['q'],args['payload']
        self.perm(self.d, u, request.url.path, args)
        jid = await self.mq.publish(q, uid, payload)
        if jid is None :
            return rest_error(self.LOG, 'APP', 'request_job', '')
        self.LOG(2, 0, label='APP', label2='request_job',  msg={'jid':jid, 'uid':uid, 'q':q})
        return rest_ok({'jid': jid})

    @requires('authenticated')
    async def get_jobs(self, request):
        u = await self.auth.check_auth(request)
        uid = u['uid']
        data = await request.json()
        kwargs = get_optional_args(data,['jstatus'])
        jobs = await self.mq.get_jobs({uid:uid,**kwargs})
        return rest_ok(jobs)

    @requires('authenticated')
    async def get_pending_jobs(self, request):
        u = await self.auth.check_auth(request)
        jobs = await self.mq.get_pending_jobs(u['uid'])
        return rest_ok(jobs)

    @requires('authenticated')
    async def get_job(self, request):
        u = await self.auth.check_auth(request)
        data = await request.json()
        [jid] = get_required(data,'jid')
        job = await self.mq.get_job(jid)
        return rest_ok(job)

    @requires('authenticated')
    async def app_route(self, request):
        if request.url.path not in self.routes:
            raise HTTPException(400, detail=f'App route not available')
        d = self.routes[request.url.path]
        f,args,kwargs = d['f'],d['args'],d['kwargs']
        u = await self.auth.check_auth(request)
        data = await request.json()
        args2 = get_required_args(data, args)
        kwargs2 = get_optional_args(data, args, kwargs)
        self.perm(self.d, u, request.url.path, {**args2, **kwargs2})
        res = await f(self.d, u, *[args2[e] for e in args2], **kwargs2)
        return rest_ok(res)



import time
import cgi
import json
from http.server import BaseHTTPRequestHandler, HTTPServer
from dl2050utils.restutils import get_required_args, get_optional_args

class Handler(BaseHTTPRequestHandler):
    
    def log_message(self, format, line, return_code, _):
        msg = f'{self.command} {return_code}'
        if self.error: msg+=f' {self.error}'
        msg += f' {self.client_address}'
        t = self.t0 if hasattr(self, 't0') else 0
        level = 2 if return_code=='200' else 4
        self.LOG(level, time.time()-t, label='WSGIServer', label2=self.path, msg=msg)

    def _set_headers(self):
        self.send_response(200)
        self.send_header('Content-type', 'application/json')
        self.end_headers()
        
    def do_HEAD(self):
        self.t0 = time.time()
        self._set_headers()
        
    def do_GET(self):
        self.t0 = time.time()
        self.send_response(400)
        self.end_headers()

    def do_POST(self):
        try:
            self.t0 = time.time()
            ctype, pdict = cgi.parse_header(self.headers.get('content-type'))
            if ctype!='application/json':
                raise Exception('Content-type error')
            if self.path not in self.routes:
                raise Exception('Invalid route')
            length = int(self.headers.get('content-length'))
            payload = json.loads(self.rfile.read(length))
            route = self.routes[self.path]
            f,args,kwargs = route['f'],route['args'],route['kwargs']
            args2 = get_required_args(payload, args)
            kwargs2 = get_optional_args(payload, args, kwargs)
            res = f(self.d, *[args2[e] for e in args2], **kwargs2)
            res = json.dumps(res)
            res = res.encode()
        except Exception as err:
            self.error = str(err)
            self.send_response(400)
            self.send_header('Content-type','application/json')
            self.end_headers()
            res = {'error': str(err)}
            res = json.dumps(res)
            res = res.encode()
            self.wfile.write(res)
            return
        self._set_headers()
        self.wfile.write(res)

def HandlerFactory(routes, d):
    class ExtendedHandler(Handler):
        def __init__(self, *args, **kwargs):
            super(ExtendedHandler, self).__init__(*args, **kwargs)
    ExtendedHandler.routes = routes
    ExtendedHandler.d = d
    ExtendedHandler.LOG = d['LOG']
    ExtendedHandler.error = None
    return ExtendedHandler

class WSGIServer():
    def __init__(self, routes, startup, d={}, secret=None, host='', port=6000):
        self.routes,self.startup = routes,startup
        self.secret,self.host,self.port = secret, host,port
        self.d = d
        self.LOG = d['LOG']
        
    def run(self, host='', port=6000):
        try:
            if self.startup(self.d):
                return True
        except Exception as err:
            self.LOG(4, 0, label='WSGIServer', label2='STARTUP EXCEPTION', msg=str(err))
            raise RuntimeError('MDB STARTUP EXCEPTION')
        Handler = HandlerFactory(self.routes, self.d)
        httpd = HTTPServer((host, port), Handler)
        self.LOG(2, 0, label='WSGIServer', label2='RUN', msg=f'MDB listenning on {self.port}')
        try:
            httpd.serve_forever()
        except KeyboardInterrupt:
            pass
        httpd.server_close()



from datetime import datetime, timezone
import bcrypt
from starlette.routing import Route
from dl2050utils.core import oget
from dl2050utils.com import send_otp
from dl2050utils.restutils import HTTPException, get_required, rest_ok, rest_error, mk_key, mk_jwt_token

NOT_REGISTERED = 0
WAIT_SET_PASSWD = 1
SUSPENDED = 2
SUSPENDED_LICENCE = 3
SUSPENDED_FAILS = 4
SUSPENDED_REQUESTS = 5
ACTIVE = 10

AUTHORIZED = 100
NOT_AUTHORIZED = 101
WRONG_PASSWD = 102
REGISTER_NOT_ALLOWED = 103
OTP_ERROR = 104
OTP_EXPIRED = 105
PASSWORD_CHANGED = 106

MAX_FAILS = 3
MAX_REQUESTS = 1e4

async def db_update(db, tbl, kc, d, prefix=''):
    res = await db.update(tbl, kc, d)
    if res:
        raise HTTPException(401, detail=f'{prefix+": " if prefix else ""}update_user db error')

def increase(d, f):
    if d is None or f not in d: return d
    if d[f] is None: d[f]=0
    d[f] += 1
    d['ts_access'] = datetime.now(timezone.utc)
    if f=='logins': d['ts_login'],d['fails'] = datetime.now(timezone.utc),0
    return d

class Auth():
    def __init__(self, cfg, LOG, NOTIFY, db):
        self.cfg, self.LOG, self.NOTIFY, self.db = cfg, LOG, NOTIFY, db

    def get_routes(self):
        return [
            Route('/api/auth/auth_info', endpoint=self.auth_info, methods=['GET']),
            Route('/api/auth/check_user', endpoint=self.check_user, methods=['POST']),
            Route('/api/auth/login', endpoint=self.login, methods=['POST']),
            Route('/api/auth/register', endpoint=self.register, methods=['POST']),
            Route('/api/auth/set_passwd', endpoint=self.set_passwd, methods=['POST']),
            Route('/api/auth/recover_passwd', endpoint=self.recover_passwd, methods=['POST']),
            Route('/api/auth/change_passwd', endpoint=self.change_passwd, methods=['POST']),
        ]

    async def startup(self):
        try:
            self.product = self.cfg['app']['product']
            self.auth_secret = self.cfg['rest']['auth_secret']
            self.allow_register = oget(self.cfg, ['rest', 'allow_register'], False)
            self.strong_passwd = oget(self.cfg, ['rest', 'strong_passwd'], False)
            self.otp_timeout = oget(self.cfg, ['rest', 'otp_timeout'], False)
            self.otp_mode = oget(self.cfg, ['rest', 'otp_mode'], False)
        except Exception as e:
            self.LOG(4, 0, label='AUTH', label2='startup', msg={'error_msg': str(e)})
            return True
        self.LOG(2, 0, label='AUTH', label2='startup', msg=f'OK')
        return False   

    async def shutdown(self):
        self.LOG(2, 0, label='AUTH', label2='shutdown', msg='OK')
        return False

    async def register_requests(self, url, uid, t, error):
        await self.db.insert('metrics', {'ts': datetime.now(timezone.utc), 'uid':uid, 'url':url, 't':t, 'error':error})
        if uid==-1: return
        user = await self.db.select_one('users', {'id': uid})
        user = increase(user, 'requests')
        if user['requests'] > MAX_REQUESTS: user['ustatus'] = SUSPENDED_REQUESTS
        await db_update(self.db, 'users', 'email', user)

    async def gen_otp(self, user, return_code=WAIT_SET_PASSWD):
        if self.otp_mode=='phone' and user['phone'] is None:
            return rest_error(self.LOG, 'AUTH', 'register', 'phone not defined')
        otp = mk_key(n=4)
        self.LOG(2, 0, label='GEN_OTP', label2='New OTP', msg={'email': user['email'], 'OTP': otp})
        err = send_otp(self.NOTIFY, self.otp_mode, self.product, user['email'], user['phone'], otp)
        if err is not None:
            self.LOG(4, 0, label='AUTH', label2='gen_otp', msg=err)
            return rest_error(self.LOG, 'AUTH', 'register', 'send key by email error')
        user['otp'],user['ts_passwd'],user['ustatus'] =  otp, datetime.now(timezone.utc),WAIT_SET_PASSWD
        await db_update(self.db, 'users', 'email', user, prefix='gen_otp')
        return rest_ok({'user_status': return_code})

    async def check_auth(self, request):
        uid = request.user.display_name
        user = await self.db.select_one('users', {'id': uid})
        if user is None:
            msg = 'invalid user'
            self.LOG(4, 0, label='AUTH', label2='check_auth', msg={'uid':uid, 'exception':msg})
            raise HTTPException(401, detail=msg)
        if user['ustatus']!=ACTIVE:
            msg = 'user not active'
            self.LOG(4, 0, label='AUTH', label2='check_auth', msg={'uid':uid, 'exception':msg})
            raise Exception('Authorization error: user not active')
        return {'uid':uid, 'email':user['email'], 'status':user['ustatus'], 'org':user['org'], 'dep':user['dep']}

    async def auth_info(self, request):
        return rest_ok({'allow_register': self.allow_register, 'strong_passwd': self.strong_passwd, 'otp_timeout': self.otp_timeout,
                        'otp_mode': self.otp_mode})

    async def check_user(self, request):
        data = await request.json()
        [email] = get_required(data,['email'])
        user = await self.db.select_one('users', {'email':email})
        if user is None:
            return rest_ok({'user_status': NOT_REGISTERED})
        if not self.allow_register and user['ustatus']==NOT_REGISTERED:
            return await self.gen_otp(user, WAIT_SET_PASSWD)
        if user['ustatus']==ACTIVE:
            user = increase(user, 'logins')
            await db_update(self.db, 'users', 'email', user)
            return rest_ok({'user_status': user['ustatus']})
        return rest_ok({'user_status':user['ustatus']})

    async def login(self, request):
        data = await request.json()
        [email,passwd] = get_required(data,['email', 'passwd'])
        user = await self.db.select_one('users', {'email':email})
        if user is None:
            self.LOG(3, 0, label='AUTH', label2='login', msg={'user':email, 'error_message':'user not registered'})
            return rest_ok({'user_status': NOT_REGISTERED})
        if user['ustatus'] in [NOT_REGISTERED, WAIT_SET_PASSWD, SUSPENDED, SUSPENDED_LICENCE]:
            self.LOG(3, 0, label='AUTH', label2='login', msg={'user':email, 'error_message': 'user not active'})
            return rest_ok({'user_status': user['ustatus']})
        if not "passwd" in user or user['passwd'] is None or passwd is None:
            return rest_error(self.LOG, 'AUTH', 'login', 'no passwd provided')
        if not bcrypt.checkpw(passwd.encode(), user['passwd'].encode()):
            user = increase(user, 'fails')
            if user['fails'] > MAX_FAILS: user['ustatus'],user['fails'] = SUSPENDED_FAILS,0
            await db_update(self.db, 'users', 'email', user)
            if user['fails'] > MAX_FAILS:
                self.LOG(4, 0, label='AUTH', label2='login', msg={'user':email, 'error_message':'user suspend - too many login fails'})
                return rest_ok({'user_status': SUSPENDED_FAILS})
            else:
                self.LOG(3, 0, label='AUTH', label2='login', msg={'user': email, 'error_message': 'wrong password'})
                return rest_ok({'user_status': WRONG_PASSWD})
        jwt_token = mk_jwt_token(user['id'], email, self.auth_secret)
        user = increase(user, 'logins')
        await db_update(self.db, 'users', 'email', user)
        return rest_ok({'user_status':AUTHORIZED, 'jwt_token':jwt_token, 'user_info':{'role':user['role'], 'org':user['org'], 'dep':user['org']}})

    async def register(self, request):
        data = await request.json()
        [email] = get_required(data,['email'])
        user = await self.db.select_one('users', {'email':email})
        if user is None:
            if not self.allow_register:
                return rest_ok({'user_status': REGISTER_NOT_ALLOWED})
            user = {'email':email, 'ustatus':NOT_REGISTERED, 'ts_insert':datetime.now(timezone.utc),
                    'logins':0, 'fails':0, 'requests':0}
            await self.db.insert('users', user)
        if user['ustatus'] in [ACTIVE, SUSPENDED, SUSPENDED_LICENCE, WAIT_SET_PASSWD]:
            return rest_ok({'user_status': user['ustatus']})
        user['ustatus'] = WAIT_SET_PASSWD
        await db_update(self.db, 'users', 'email', user)
        return await self.gen_otp(user, WAIT_SET_PASSWD)

    async def recover_passwd(self, request):
        data = await request.json()
        [email] = get_required(data,['email'])
        user = await self.db.select_one('users', {'email':email})
        if user is None:
            return rest_error(self.LOG, 'AUTH', 'SERVICE_ERROR', 'recover_passwd: user not found')
        user['ustatus'] = WAIT_SET_PASSWD
        await db_update(self.db, 'users', 'email', user)
        await self.gen_otp(user, WAIT_SET_PASSWD)
        return rest_ok({})

    async def set_passwd(self, request):
        data = await request.json()
        [otp,passwd] = get_required(data,['otp','passwd'])
        user = await self.db.select_one('users', {'otp': otp})
        if user is None:
            return rest_ok({'user_status': OTP_ERROR})
        delta = int(datetime.now(timezone.utc).strftime("%s")) - int(user['ts_passwd'].strftime("%s"))
        if delta > self.otp_timeout:
            self.LOG(3, 0, label='AUTH', label2='OTP', msg=f'OTP expiered - delta={delta}')
            return await self.gen_otp(user, OTP_EXPIRED)
        hashed = bcrypt.hashpw(passwd.encode(), bcrypt.gensalt())
        user['passwd'],user['ts_passwd'],user['otp'],user['ustatus'] = hashed.decode(),datetime.now(timezone.utc),'',ACTIVE
        increase(user, 'logins')
        await db_update(self.db, 'users', 'email', user)
        jwt_token = mk_jwt_token(user['id'], user['email'], self.auth_secret)
        return rest_ok({'user_status': AUTHORIZED, 'jwt_token': jwt_token})

    async def change_passwd(self, request):
        uid = request.user.display_name
        if uid is None or uid=='':
            return rest_error(self.LOG, 'AUTH', 'change_passwd', 'User ID not set')
        user = await self.db.select_one('users', {'id': uid})
        if user is None:
            return rest_error(self.LOG, 'AUTH', 'SERVICE_ERROR', 'change_passwd: user not registered')
        data = await request.json()
        [passwd,new_passwd] = get_required(data,['passwd','new_passwd'])
        if not "passwd" in user or user['passwd'] is None or passwd is None:
            return rest_error(self.LOG, 'AUTH', 'SERVICE_ERROR', 'change_passwd: password error')
        if not bcrypt.checkpw(passwd.encode(), user['passwd'].encode()):
            return rest_ok({'user_status': WRONG_PASSWD})
        hashed = bcrypt.hashpw(new_passwd.encode(), bcrypt.gensalt())
        user['passwd'],user['otp'],user['ustatus'] = hashed.decode(),'',ACTIVE
        await db_update(self.db, 'users', 'email', user)
        return rest_ok({'user_status': PASSWORD_CHANGED})



import os
from pathlib import Path
from dotenv import load_dotenv
import yaml

def find_yml(name):
    p = Path(f'./{name}')
    if p.exists(): return p
    p = Path(f'/run/secrets/{name}')
    if p.exists(): return p
    return None

def config_load(name=None):
    load_dotenv()
    name = 'config.yml' if name is None else f'config-{name}.yml'
    fname = find_yml(name)
    if fname is None:
        raise RuntimeError('Unable to find config file')
    with open(str(fname),'r') as stream:
        try:
            cfg = yaml.safe_load(stream)
        except yaml.YAMLError as err:
            raise RuntimeError(f'Unable to load config file: {str(err)}')
    return cfg

def get_db_url(Config, with_db_name=False):
    if Config is None:
        return None
    if not 'db' in Config:
        return None
    if os.getenv('development') is None:
        if not 'host' in Config['db']:
            return None
        host = Config['db']['host']
    else:
        if not 'host_dev' in Config['db']:
            return None
        host = Config['db']['host_dev']
    if not 'port' in Config['db']:
        port = 3306
    else:
        port = Config['db']['port']
    if not 'name' in Config['db']:
        return None
    if not 'passwd' in Config['db']:
        return None
    passwd = Config['db']['passwd']
    url = f'mysql+pymysql://root:{passwd}@{host}:{port}'
    if with_db_name:
        name = Config['db']['name']
        url = f'{url}/{name}'
    return url

def get_rest_url(Config):
    if Config is None:
        return None
    if not 'rest' in Config:
        return None
    if os.getenv('development') is None:
        if not 'host' in Config['rest']:
            return None
        host = Config['rest']['host']
    else:
        if not 'host_dev' in Config['rest']:
            return None
        host = Config['rest']['host_dev']
    if not 'port' in Config['rest']:
        return None
    port = Config['rest']['port']
    url = f'{host}:{port}'
    return url


import datetime
import random
import json
import hashlib
import jwt
from starlette.responses import JSONResponse
import orjson

class HTTPException(Exception):
    def __init__(self, status_code, detail:None):
        if detail is None: detail = ''
        self.status_code = status_code
        self.detail = detail

    def __repr__(self) -> str:
        class_name = self.__class__.__name__
        return f"{class_name}(status_code={self.status_code!r}, detail={self.detail!r})"

def get_required_args(payload, args):
    args2,miss = {},[]
    for e in args:
        if e not in payload or payload[e] is None:
            miss.append(e)
        else:
            args2[e] = payload[e]
    if len(miss):
        raise HTTPException(400, f'Missing required args: {", ".join(miss)}')
    return args2

def get_optional_args(payload, required, args):
    optional = {e:payload[e] for e in payload if e not in required}
    invalid = []
    for e in optional:
        if e not in args:
            invalid.append(e)
    if len(invalid):
        raise HTTPException(400, f'Invalid optional args: {", ".join(invalid)}')
    return optional

def get_required(payload, args):
    args2 = []
    for e in args:
        if e not in payload or payload[e] is None:
            raise HTTPException(400, detail=f'Missing required arg {e}')
        args2.append(payload[e])
    return args2

class OrjsonResponse(JSONResponse):
    def render(self, content):
        return orjson.dumps(content)

def mk_key(n=4):
    return ''.join([chr(48+i) if i<10 else chr(65-10+i) for i in [random.randint(0, 26+10-1) for _ in range(n)]])
    # return ''.join(random.choice(string.ascii_lowercase) for i in range(n))

def get_hash(o, secret):
    o['secret']=secret
    return hashlib.sha224(json.dumps(o).encode()).hexdigest()

def check_hash(o, h, secret):
    o['secret']=secret
    return get_hash(o, secret)==h

def mk_jwt_token(uid, email, secret):
    JWT_EXP_DELTA_SECONDS = 30*24*3600
    payload = { 'uid': uid, 'email': email, 'username': '', 'exp': datetime.datetime.utcnow() + datetime.timedelta(seconds=JWT_EXP_DELTA_SECONDS)}
    return jwt.encode(payload, secret, 'HS256') # .decode('utf-8')

def rest_ok(result):
    return OrjsonResponse({'status': 'OK', 'result': result})

def rest_error(LOG, label, label2, error_msg):
    LOG(4, 0, label=label, label2=label2, msg=error_msg)
    return OrjsonResponse({'status': 'ERROR', 'error_msg': error_msg})

async def get_meta(db, model):
    row = await db.select_one('models', {'model': model})
    if row is not None: return json.loads(row['meta'])
    return None

def sync_get_meta(db, model):
    row = db.sync_select_one('models', {'model': model})
    if row is not None: return json.loads(row['meta'])
    return None

def mk_weeks(ds1='2018-01-01', ds2=None, weekday=6):
    d1 = datetime.datetime.strptime(ds1, '%Y-%m-%d').date()
    delta = 5 - d1.weekday()
    if delta<0: delta+=7
    d1 += datetime.timedelta(days=delta)
    d2 = datetime.datetime.now().date() if ds2 is None else datetime.datetime.strptime(ds2, '%Y-%m-%d').date()
    ds = [d.strftime("%Y-%m-%d") for d in rrule.rrule(rrule.WEEKLY, dtstart=d1, until=d2)]
    return ds[::-1]

def get_week2(weeks, week): return weeks[weeks.index(week)+1] if weeks.index(week)+1<len(weeks) else None

def s3_urls(s3, bucket_name, prefix):
    response = s3.list_objects_v2(Bucket=bucket_name, Prefix=prefix, MaxKeys=1024)['Contents']
    return [f'http://{bucket_name}.s3-eu-west-1.amazonaws.com/{e["Key"]}' for e in response]



import time
import uvicorn
from starlette.applications import Starlette
from starlette.responses import JSONResponse
from starlette.middleware import Middleware
from starlette.middleware.base import BaseHTTPMiddleware
from starlette.middleware.cors import CORSMiddleware
from starlette_jwt import JWTAuthenticationBackend
from starlette.middleware.authentication import AuthenticationMiddleware
from starlette.responses import JSONResponse
from starlette.routing import Mount
from starlette.staticfiles import StaticFiles
from dl2050utils.core import oget
from dl2050utils.env import config_load
from dl2050utils.log import AppLog
from dl2050utils.com import Notify, read_address_book
from dl2050utils.db import DB
from dl2050utils.mq import MQ
from dl2050utils.auth import Auth
from dl2050utils.restapp import App, HTTPException

class ASGI_Server():

    def __init__(self, name, path, routes, appstartup, perm, qs):
        cfg = config_load(name)
        LOG = AppLog(cfg)
        NOTIFY = Notify(cfg=cfg, address_book=read_address_book())
        dbname = oget(cfg, ['db','dbname'], 'postgres')
        db = DB(cfg=cfg, log=LOG, dbname=dbname)
        mq = MQ(LOG, db, qs)
        auth = Auth(cfg, LOG, NOTIFY, db)
        app = App(cfg, LOG, NOTIFY, db, mq, auth, path, routes, appstartup, perm)
        self.LOG,self.db,self.mq,self.auth,self.app = LOG,db,mq,auth,app
        self.port = oget(cfg, ['rest','port'], 5000)
        self.exception_handlers = {
            HTTPException: self.http_exception,
            404: self.http_exception,
            Exception: self.server_error_exception,
            500: self.server_error_exception,
        }
        auth_secret = cfg['rest']['auth_secret']
        self.middleware = [
            Middleware(RestStartMiddleware),
            Middleware(CORSMiddleware, allow_origins=['*'], allow_methods=['*'], allow_headers=['*'], expose_headers=['*']),
            Middleware(AuthenticationMiddleware, backend=JWTAuthenticationBackend(secret_key=auth_secret, prefix='Bearer',
                       username_field='uid')),
            Middleware(RestEndMiddleware, LOG=self.LOG, auth=auth),
        ]
        self.routes = auth.get_routes()+app.get_routes()
        
    async def startup(self):
        if await self.db.startup(): raise('DB startup error')
        if await self.mq.startup(): self.LOG(3, 0, label='REST', label2='STARTUP', msg='MQ not available')
        if await self.auth.startup(): raise('Auth startup error')
        if await self.app.startup(): raise('App startup error')
        self.LOG(2, 0, label='REST', label2='STARTUP', msg=f'OK (port {self.port})')
        return False
            
    async def shutdown(self):
        self.LOG(2, 0, label='REST', label2='SHUTDOWN')
    
    async def http_exception(self, request, exc):
        request.state.error = True
        request.state.error = exc.detail
        return JSONResponse({'status': 'HTTP_EXCEPTION', 'error_msg': exc.detail}, status_code=exc.status_code)

    async def server_error_exception(self, request, exc):
        self.LOG(4, 0, label='REST', label2='SERVER_EXCEPTION', msg=exc.detail)
        return JSONResponse({'status': 'SERVER_EXCEPTION', 'error_msg': exc.detail}, status_code=exc.status_code)
        
    def run(self, port=None):
        rest = Starlette(
            debug=True,
            exception_handlers=self.exception_handlers,
            middleware=self.middleware,
            on_startup=[self.startup],
            on_shutdown=[self.shutdown],
            routes=self.routes,
        )
        port = port or self.port
        uvicorn.run(rest, port=port, host='0.0.0.0', log_level='critical')

class RestStartMiddleware(BaseHTTPMiddleware):
    def __init__(self, app):
        super().__init__(app)
    async def dispatch(self, request, call_next):
        request.state.error = False
        request.state.t0 = time.time()
        return await call_next(request)

class RestEndMiddleware(BaseHTTPMiddleware):
    def __init__(self, app, LOG=None, auth=None):
        super().__init__(app)
        self.LOG,self.auth=LOG,auth
    async def dispatch(self, request, call_next):
        response = await call_next(request)
        url,uid,error,t = request.url.path,request.user.display_name,request.state.error,time.time()-request.state.t0
        if uid=='': uid=-1
        if request.state.error:
            self.LOG(4, t, label='REST', label2='HTTP_EXCEPTION', msg=request.state.error)
            return response
        if request.url.path not in ['/api/auth/is_auth', '/api/get_meta']:
            await self.auth.register_requests(url, uid, t, error)
        self.LOG(2, t, label='REST', label2=request.url.path, msg={'uid':uid})
        return response



import datetime
import logging
from dl2050utils.core import oget

LOG_FORMAT = '\
%(name)s  %(levelname)s - %(asctime)s - %(product)s:%(client)s:%(instance)s:%(service)s - \
%(label)s - %(label2)s - %(duration).3f - %(message)s'

def parse_msg(d):
    if d is None: return 'null'
    if type(d) in [int, float]: return str(d)
    if isinstance(d, str): return d.replace('\n', ' ')
    if isinstance(d, datetime.datetime): return d.strftime("%Y-%m-%Y %H:%M:%S")
    if type(d) is list: return f'[#{len(d)}]'
    if type(d) is not dict: return f'OBJECT-{type(d)}'
    s = '{'
    for key, val in d.items():
        if isinstance(d[key], dict):
            s += f'{key}:'+'{...}'
        elif type(d[key]) is list:
            s += f'{key}:[#{len(d)}]'
        else:
            s += f'{key}:{val}'
        s += ', '
    return s[:-2]+'}'

class AppLog():
    def __init__(self, cfg, service='rest'):
        global loggers
        self.msg_prefix = {
            'product': oget(cfg, ['app','product'], '_PRODUCT_'),
            'client': oget(cfg, ['app','client'], '_CLIENT_'),
            'instance': oget(cfg, ['app','instance'], '_INSTANCE_'),
            'service': service,
        }
        logging.basicConfig(level=logging.INFO)
        logger = logging.getLogger('APPLOG')
        logger.propagate = False
        while len(logger.handlers):
            logger.removeHandler(logger.handlers[0])
        streamHandler = logging.StreamHandler()
        streamHandler.setLevel(logging.INFO)
        formatter = logging.Formatter(LOG_FORMAT)
        streamHandler.setFormatter(formatter)
        logger.addHandler(streamHandler)
        self.LOG_LEVELS = {
            1: logger.debug,
            2: logger.info,
            3: logger.warning,
            4: logger.error,
            5: logger.critical
        }
        
    def __call__(self, level=3, t=None, label=None, label2=None, msg=''):
        if not isinstance(level, int) or level<1 or level>5: return
        log_f = self.LOG_LEVELS[level]
        log_msg = self.msg_prefix
        log_msg['label'],log_msg['label2'],log_msg['duration'] = label,label2,t or 0
        s = parse_msg(msg)
        log_f(s, extra=log_msg)

# class ServiceLog():
#     def __init__(self, server_name='WSGI'):
#         self.LEVELS = {1:'DEBUG', 2:'INFO', 3:'WARNING', 4:'ERROR', 5:'CRITICAL'}
#         self.server_name = server_name
#     def __call__(self, level, t, label='', label2='', msg=''):
#         print(f'{self.server_name} - {datetime.datetime.now().isoformat()} - {self.LEVELS[level]} - {label} - {label2} - {t:.3f} - {msg}')

class BaseLog():
    def log(self, level, t, label=None, label2=None, msg=None):
        print(f'{level}:{t}:{label}:{label2}:{msg}')



#
# draft
#

TBLS_EXCLUDE = ['refs', 'trans', 'enums', 'information_schema_catalog_name', 'check_constraint_routine_usage', 'applicable_roles',
                'administrable_role_authorizations','collation_character_set_applicability', 'attributes', 'check_constraints',
                'character_sets', 'collations', 'column_domain_usage', 'column_column_usage', 'column_privileges',
                'column_udt_usage', 'columns', 'constraint_column_usage', 'schemata', 'constraint_table_usage',
                'domain_constraints', 'sql_packages', 'domain_udt_usage', 'sequences', 'domains', 'enabled_roles',
                'key_column_usage', 'parameters', 'referential_constraints', 'sql_features', 'role_column_grants',
                'routine_privileges', 'role_routine_grants', 'routines', 'sql_implementation_info', 'sql_parts',
                'sql_languages', 'sql_sizing', 'sql_sizing_profiles', 'table_constraints', 'table_privileges',
                'role_table_grants', 'views', 'tables', 'transforms', 'triggered_update_columns', '_pg_foreign_servers',
                'triggers', 'data_type_privileges', 'udt_privileges', 'role_udt_grants', 'usage_privileges', 'element_types',
                'role_usage_grants', 'user_defined_types', '_pg_foreign_table_columns', 'view_column_usage', 'view_routine_usage',
                'view_table_usage', 'foreign_server_options', 'column_options', '_pg_foreign_data_wrappers',
                'foreign_data_wrapper_options', 'foreign_tables', 'foreign_data_wrappers', 'foreign_servers',
                'foreign_table_options', 'user_mappings', 'user_mapping_options']

DRILLS = {
    'tbls': {'canal': ['zona', 'subzona', 'agente']}
}

TYPES = {
    'character varying': 'S',
    'character': 'S',
    'integer': 'I',
    'real': 'P',
    'double precision': 'F',
    'money': 'C',
    'date': 'D',
    'timestamp with time zone': 'T',
    'boolean': 'B',
    'ARRAY': 'ARRAY',
    'json': 'JSON',
    'USER-DEFINED': 'ENUM'
}

Q_DROP = """
    DROP TABLE flds;
    DROP TABLE enums;
    DROP TABLE refs;
    DROP TABLE tbls;
    DROP TYPE aligns;
"""

Q_CREATE = """
    CREATE TYPE aligns AS ENUM ('Left', 'Center', 'Right');
    
    CREATE TABLE tbls (
        tbl varchar(64) primary key,
        name varchar(64),
        keys TEXT [],
        descf varchar(64),
        drills json
    );

    CREATE TABLE flds (
        tbl varchar(64) NOT NULL references tbls(tbl),
        fld varchar(64) NOT NULL,
        name varchar(64),
        type varchar(64),
        size char(16),
        frmt char(16),
        ronly boolean,
        align aligns,
        PRIMARY KEY(tbl, fld)
    );
    
    CREATE TABLE refs (
        tbl varchar(64) NOT NULL references tbls(tbl),
        fld varchar(64) NOT NULL,
        tbl_ref varchar(64) NOT NULL references tbls(tbl),
        fld_ref varchar(64) NOT NULL
    );
    
    CREATE TABLE enums (
        name varchar(64) NOT NULL,
        values TEXT []
    );
    
"""

Q_TBLS = """
    SELECT table_name FROM information_schema.tables;
"""

Q_FLDS_ = f"""
    SELECT table_name as tbl, column_name as fld, column_name as name, data_type as type, character_maximum_length as size
    FROM INFORMATION_SCHEMA.COLUMNS
    WHERE table_name IN 
"""

Q_KEYS = """
    SELECT tc.table_name as tbl, c.column_name as fld
    FROM information_schema.table_constraints tc 
    JOIN information_schema.constraint_column_usage AS ccu USING (constraint_schema, constraint_name) 
    JOIN information_schema.columns AS c ON c.table_schema = tc.constraint_schema
        AND tc.table_name = c.table_name AND ccu.column_name = c.column_name
    WHERE constraint_type = 'PRIMARY KEY';
"""

Q_REFS = """
    SELECT
        tc.table_name as tbl, 
        kcu.column_name as fld, 
        ccu.table_name AS tbl_ref,
        ccu.column_name AS fld_ref
    FROM 
        information_schema.table_constraints AS tc 
        JOIN information_schema.key_column_usage AS kcu
          ON tc.constraint_name = kcu.constraint_name
          AND tc.table_schema = kcu.table_schema
        JOIN information_schema.constraint_column_usage AS ccu
          ON ccu.constraint_name = tc.constraint_name
          AND ccu.table_schema = tc.table_schema
    WHERE tc.constraint_type = 'FOREIGN KEY' and tc.constraint_name ='flds_tbl_fkey';
"""

Q_ENUMS = """
    select t.typname as ename,  
           e.enumlabel as evalue
    from pg_type t 
       join pg_enum e on t.oid = e.enumtypid  
       join pg_catalog.pg_namespace n ON n.oid = t.typnamespace;
"""
async def build_meta(db):
    for q in Q_DROP.split(";"): await db.query(q)
    for q in Q_CREATE.split(";"): await db.query(q)
    
    tbls = await db.query(Q_TBLS)
    tbls_list = [e['table_name'] for e in tbls
            if e['table_name'] not in TBLS_EXCLUDE and not e['table_name'].startswith('pg') and not e['table_name'].startswith('_pg')]
    keys = await db.query(Q_KEYS)
    tbls = [{'tbl': e, 'name': e, 'keys': [e1['fld'] for e1 in keys if e1['tbl']==e], 'drills': DRILLS[e] if e in DRILLS else {}}
            for e in tbls_list]
    await db.insert_rows('tbls', tbls, delete=False)
    
    TBLS_LIST = ', '.join([f'\'{e}\'' for e in tbls_list])
    Q_FLDS = f'{Q_FLDS_} ({TBLS_LIST});'
    flds = await db.query(Q_FLDS)
    for fld in flds:
        fld['name'] = fld['name'].title()
        fld['type'] = TYPES[fld['type']]
        fld['align'] = 'Left' if fld['type'] in ['S', 'D', 'T'] else 'Right'
    await db.insert_rows('flds', flds, delete=True)
        
    enums = await db.query(Q_ENUMS)
    enums = [{'name': e, 'values': [e1['evalue'] for e1 in enums if e1['ename']==e]} for e in set([e['ename'] for e in enums])]
    await db.insert_rows('enums', enums, delete=True)
    
    refs = await db.query(Q_REFS)
    await db.insert_rows('refs', refs, delete=True)

async def update_meta(db, meta):
    for tbl in meta['tables']:
        if 'fields' not in meta['tables'][tbl]: continue
        for fld in meta['tables'][tbl]['fields']:
            d = {'tbl': tbl, 'fld': fld}
            for attr in meta['tables'][tbl]['fields'][fld]:
                d[attr] = meta['tables'][tbl]['fields'][fld][attr]
            res = await db.update('flds', ['tbl', 'fld'], d)

async def get_meta(db):
    tbls = await db.select('tbls')
    if tbls is None: return None
    flds = await db.select('flds')
    if flds is None: return None
    meta = {'tables': {e['tbl']: {'table': e, 'fields': {e1['fld']: e1 for e1 in flds if e1['tbl']==e['tbl']}} for e in tbls}}
    return meta


import os, re
import socket
import json
import base64
from sendgrid import SendGridAPIClient
from sendgrid.helpers.mail import (Mail, Attachment, FileContent, FileName, FileType, Disposition)
from twilio.rest import Client
from email.mime.text import MIMEText
from email.mime.multipart import MIMEMultipart
from email.mime.base import MIMEBase
from email import encoders
from multiprocessing import Process
import boto3
import mimetypes
mimetypes.init()
from dl2050utils.core import *

default_address_book = {
    'jneto': {'email': 'joao.filipe.neto@gmail.com', 'phone': '+351966221506'},
    'pneto': {'email': 'pneto.98@gmail.com', 'phone': '+351910528813‬'},
    'jranito': {'email': 'joao.vasco.ranito@gmail.com', 'phone': '+351966221505'},
    'rassuncao': {'email': 'rui.assuncao@ndrive.com ', 'phone': '+351933339307'}
}

def read_address_book():
    try:
        with open('./.addressbook') as f: d = json.loads(f.read())
    except IOError:
        return None
    return d

# SendGrid
def send_mail(api_key, to, subject=None, msg=None, html=None, files=[]):
    to = listify(to)
    try:
        msg = msg or '   '
        html2 = html or msg
        subject = subject or '  '
        msg = Mail(from_email='ops@dl2050.com', to_emails=to, subject=subject, html_content=html2)
        if len(files):
            attachments = []
            for file in files:
                with open(str(file), 'rb') as f: data=f.read()
                data = base64.b64encode(data).decode()
                attachedFile = Attachment(
                    FileContent(data),
                    FileName(file.name),
                    FileType(mimetypes.types_map[file.suffix]),
                    Disposition('attachment')
                )
                attachments.append(attachedFile)
            msg.attachment = attachments
        sg = SendGridAPIClient(api_key)
        res = sg.send(msg)
    except Exception as e:
        return f'SendGrid ERROR: {str(e)}'
    if res.status_code!=202: return f'SendGrid ERROR: status code={res.status_code}'
    return None

def send_mail_async(api_key, to, subject=None, msg=None, html=None, files=[]):
    p = Process(target=send_mail, args=(api_key, to, subject, msg, html, files), daemon=True)
    p.start()

# def send_sms_aws(sms_id, sms_passwd, to, msg):
#     MessageAttributes={'AWS.SNS.SMS.SenderID': {'DataType': 'String','StringValue': 'DLOPS'}}
#     client = boto3.client('sns', aws_access_key_id=sms_id, aws_secret_access_key=sms_passwd, region_name="eu-west-1")
#     client.publish(PhoneNumber=to, Message=msg, MessageAttributes=MessageAttributes)
#     return None

def send_sms(account_sid, auth_token, service_name, to, msg):
    to = listify(to)
    try:
        client = Client(account_sid, auth_token)
        for e in to:
            message = client.messages.create(body=msg, from_=service_name, to=e)
        return None
    except Exception as e:
        return f'Twilio Exception: {str(e)}'

class Notify():
    def __init__(self, cfg=None, api_key=None, sms_id=None, sms_passwd=None, address_book=None):
        self.address_book = address_book if address_book is not None else default_address_book
        if cfg is not None:
            try:
                api_key = cfg['email']['sendgrid_api_key']
                sms_id = cfg['aws']['aws_access_key_id']
                sms_passwd = cfg['aws']['aws_secret_access_key']
            except Exception as e:
                print(f'Config ERROR: cant find variable: {e}')
        self.api_key = api_key
        self.sms_id = sms_id
        self.sms_passwd = sms_passwd
        
    def __call__(self, how, to, subject=None, msg=None, html=None, files=[]):
        if how not in ['email', 'email_async', 'sms']: return 'Invalid method, options are email, email_async or sms'
        # if isinstance(ini_list1, list)
        to = listify(to)
        if not len(to): return 'No destination addresses'
        for e in to:
            if e not in self.address_book: return f'Destination {e} not found in address book'
        if how=='email' or how=='email_async':
            if self.api_key is None: return 'email credentials not defined'
            to = [self.address_book[e]['email'] for e in to]
        if how == 'email_async':
            return self.send_mail_async(to, subject=subject, msg=msg, html=html, files=files)
        if how == 'email':
            return self.send_mail(to, subject=subject, msg=msg, html=html, files=files)
        if how=='sms':
            if self.sms_id is None or self.sms_passwd is None: return 'sms credentials not defined'
            to = [self.address_book[e]['phone'] for e in to]
            return send_sms(self.sms_id, self.sms_passwd, to, msg)
    
    def send_mail_async(self, to, subject=None, msg=None, html=None, files=[]):
        send_mail_async(self.api_key, to, subject=subject, msg=msg, html=html, files=files)
    
    def send_mail(self, to, subject=None, msg=None, html=None, files=[]):
        return send_mail(self.api_key, to, subject=subject, msg=msg, html=html, files=files)

    def send_sms(self, to, msg):
        return send_sms(self.sms_id, self.sms_passwd, to, msg)

EMAIL_TEMPLATE = \
"""
<html>
<head>
    <link href="https://fonts.googleapis.com/css?family=Muli::100,200,300,400,500,600,700,800" rel="stylesheet">
</head>
    <body style="position: relative; float: left; width: 100%; height: 100%;  text-align: center; font-family: 'Muli', sans-serif;">
        <h2 style="float: left; width: 100%; margin: 40px 0px 10px 0px; font-size: 16px; text-align: center; color: #555555;">{msg}</h2>
        <h2 style="float: left; width: 100%; margin: 0px 0px 40px 0px; font-size: 24px; text-align: center; color: #61C0DF; font-weight: bold;">{otp}</h2>
    </body>
</html>
"""

def send_otp_by_email(notify, product, email, otp):
    try:
        subject = f'{product} OTP'
        msg = f'{product} OTP: '
        html = EMAIL_TEMPLATE
        html = re.sub(r'{msg}', msg, html)
        html = re.sub(r'{otp}', f'{otp}', html)
        notify.send_mail_async(email, subject=subject, html=html)
    except Exception as e:
        return str(e)
    return None

def send_otp_by_phone(notify, product, phone, otp):
    msg = f'{product} OTP: {otp}'
    try:
        notify.send_sms(phone, msg)
    except Exception as e:
        return str(e)
    return None

def send_otp(notify, mode, product, email, phone, otp):
    if mode=='phone': return send_otp_by_phone(notify, product, phone, otp)
    return send_otp_by_email(notify, product, email, otp)


def get_ip():
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    try:
        s.connect(('10.255.255.255', 1))
        IP = s.getsockname()[0]
    except:
        IP = '127.0.0.1'
    finally:
        s.close()
    return IP